// pxscan -- linux driver for the Primax Colorado Direct 9600 scanner
//
// Author: Gustaf Krling <g.kroling@swipnet.se>
//         This driver was built from the primax_scan-0.2 driver. See below.
//         Many thanks to the people below.
//
//   Copyright (C) 2000 Gustaf Krling
/*
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */        


/* primax_scan -- linux driver for Primax scanners

   Authors:
   Marco Foglia <Marco.Foglia@switzerland.org>
   Christian Ordig <chr.ordig@gmx.net>
   Thomas Schano <schano@t-online.de>

   Copyright (C) 1999 Marco Foglia, Christian Ordig, Thomas Schano

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */        


#define VERSION "0.41"

// 0.1 000324: Gustaf Kroling <g.kroling@swipnet.se>
//             - Initial Version based mostly on version 0.2 of primax_scan
//
// 0.2 000517: Gustaf Krling
//             - This change is based on a patch from Christian Ordig <chr.ordig@gmx.net>:
//               wake_up2 modified for switching off line. This reopens the connection
//               with the printer.
//	   
//             - Lots of unused code removed. ( But not all. )
//               This should remove the 'gamma' problem during linking.
//               Don't compile with -Wall switch if you want a clean compile report.
//
//             - Handling of resolution better understood. It can be set to any 
//               integer value in the range 2 to 300 inclusive.
//               Thanks to Peter Popovec <popovec@oko.fei.tuke.sk>
//
//             - Some comments added.
//
//             - This change is based on a patch from Christian Ordig <chr.ordig@gmx.net>:
//               The scanner is probed for ASIC type and scanner model.
//
// 0.3	       - This level skipped to avoid confusion with the 0.3prel version from  Peter Popovec <popovec@oko.fei.tuke.sk>
//
// 0.4	       Gustaf Krling
//             - More code removed. Some code rewritten.
//               Faster code thanks to Peter Popovec <popovec@oko.fei.tuke.sk>
//
//             - Y-offset as given in parameter -p added to start position.
//
//             - Faster movements.
//               Thanks to Peter Popovec <popovec@oko.fei.tuke.sk>
//
//             - B/W mode added. Uses the red channel.
//
//             - Gray mode added.
//
//             - Manual start added. Waits for scanner button to be pressed before scanning starts.
//
//	       - Calibration of individual pixels in x-direction fixed.
//
//	       - Wrap around in gamma calculation fixed.
// 
// 0.41	010304: Gustaf Krling
//             - Minor fixes. No new functionality.
//		
#include <asm/io.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <unistd.h>            
#include <sched.h>
#include <sys/time.h>
#include <tiffio.h>
#include <getopt.h>
#include <unistd.h>
#include <string.h>

#include "capability.h"                        
#include "pxscan.h"
#include "registers.h"

//#define PORT 0x378;
                             

static const int x_table_size = 0x9F6; 
static const int y_table_size = 0xBDD;
static const int calibration_table_size = 0x1DE2; 
static const int color_table_size = 0x1000;
static const int num_buffers = 12;


//static int dbg_level = DBG_HIGH;
static int dbg_level = DBG_TIMEOUT;


//#define SOFTEPP		//*gk not tested in this version

int redmax,greenmax,bluemax,redmin,greenmin,bluemin;
int *whiteline,redlow,greenlow,bluelow,redgain,greengain,bluegain;
int quietflag=0;



// debugging code here

void dbgstop( int no )
{
	printf("dbgstop %i \n",no);
	no=getchar();
}

/*-------------------------------------------------------------*/
void epp_off(void) ;

/*-------------------------------------------------------------*/
void epp_on(void);

/*-------------------------------------------------------------*/
/*
 * resets a timer
 */
inline void init_timer(long long *timer, unsigned long us);

  
/*-------------------------------------------------------------*/
/*
 * checks for timeout 
 */
inline int check_timer(long long *timer) ;


/*-------------------------------------------------------------*/
int clear_timeout(void);

/*-------------------------------------------------------------*/
/* 
 * Sets mode of next read/write operation from/to BASE+4.  
 * Does nothing if already in right mode
 */
int epp_set_mode(unsigned char mode); 



/*-------------------------------------------------------------*/
/*
 * switching the scanner online
 */

/* Switch the scanner online */    
 unsigned char switcher_online (void);


/* Switch scanner offline - back to printer pass thru */
void switcher_offline (void);

/*-------------------------------------------------------------*/
//   programmed input routine

unsigned char pgmin();


int epp_read(unsigned char *data, int num, unsigned char mode) ;

/*-------------------------------------------------------------*/

// programmed output routine

inline void pgmout(unsigned char databyte);


int epp_write(unsigned char *data, int num, unsigned char mode) ;

/*-------------------------------------------------------------*/
/* 
 * calculate the color matrix
 *
 */
void calc_color_table (scan_image *image)
{
    int i;
    double midin, val, m, b;
    int maxin = 4095;
    int maxout = 255;


// do the gamma correction observing the actual range of input data 

	midin=(double)(redmin+redmax)/2.0;
	b = (1.0 + image->color[0].brightness / 100.0) * midin;
	m = 1.0 + image->color[0].contrast / 100.0;
	for (i=0; i <= maxin; i++){
	    val=((double)i-midin)*m+b;
	    if((int)val<redmin) image->color_table[0][i] = 0;
	    else {
		if((int)val>redmax) image->color_table[0][i] = 0xff;
		else {
	            image->color_table[0][i] = maxout * pow( (val-(double)redmin)/(double)(redmax-redmin),(1.0 / image->color[0].gamma) );
		}
	    }
	}

	midin=(double)(greenmin+greenmax)/2.0;
	b = (1.0 + image->color[1].brightness / 100.0) * midin;
	m = 1.0 + image->color[1].contrast / 100.0;
	for (i=0; i <= maxin; i++){
	    val=((double)i-midin)*m+b;
	    if((int)val<greenmin) image->color_table[1][i] = 0;
	    else {
		if((int)val>greenmax) image->color_table[1][i] = 0xff;
		else {
	            image->color_table[1][i] = maxout * pow( (val-(double)greenmin)/(double)(greenmax-greenmin),(1.0 /image->color[1].gamma) );
		}
	    }
	}

	midin=(double)(bluemin+bluemax)/2.0;
	b = (1.0 + image->color[2].brightness / 100.0) * midin;
	m = 1.0 + image->color[2].contrast / 100.0;
	for (i=0; i <= maxin; i++){
	    val=((double)i-midin)*m+b;
	    if((int)val<bluemin) image->color_table[2][i] = 0;
	    else {
		if((int)val>bluemax) image->color_table[2][i] = 0xff;
		else {
	            image->color_table[2][i] = maxout * pow( (val-(double)bluemin)/(double)(bluemax-bluemin),(1.0 / image->color[2].gamma));
		}
	    }
	}

	DBG(DBG_MEDIUM,"min= %i %i %i  max= %i %i %i \n",redmin,greenmin,bluemin,redmax,greenmax,bluemax);

}
/*-------------------------------------------------------------*/
/* 
 * opens a tiff file for writing
 */
void tiff_open(TIFF **tiffile, scan_image *image)
{
    *tiffile = TIFFOpen(image->filename, "w");
    if (!tiffile) {
        DBG(DBG_HELPERS, "Error during save: tiff_open did not open file \n");
        return;
    }
    
    TIFFSetField(*tiffile, TIFFTAG_IMAGEWIDTH, image->width);
    TIFFSetField(*tiffile, TIFFTAG_IMAGELENGTH, (image->height)-1);		// skip first line
    TIFFSetField(*tiffile, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    TIFFSetField(*tiffile, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField(*tiffile, TIFFTAG_ROWSPERSTRIP,
    TIFFDefaultStripSize(*tiffile, -1));


    switch( image->color_mode)
    {
        case  RGB_MODE:
            TIFFSetField(*tiffile, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
            TIFFSetField(*tiffile, TIFFTAG_SAMPLESPERPIXEL, 3);
            TIFFSetField(*tiffile, TIFFTAG_BITSPERSAMPLE, 8);
            TIFFSetField(*tiffile, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
            break;
        case  GRAY_MODE:
            TIFFSetField(*tiffile, TIFFTAG_PHOTOMETRIC, 1);
            TIFFSetField(*tiffile, TIFFTAG_SAMPLESPERPIXEL, 1);
            TIFFSetField(*tiffile, TIFFTAG_BITSPERSAMPLE, 8);
            TIFFSetField(*tiffile, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
            break;
        case  BW_MODE:
        case  BW_MODE_RED:
        case  BW_MODE_GREEN:
        case  BW_MODE_BLUE:
            TIFFSetField(*tiffile, TIFFTAG_PHOTOMETRIC, 1);
            TIFFSetField(*tiffile, TIFFTAG_SAMPLESPERPIXEL, 1);
            TIFFSetField(*tiffile, TIFFTAG_BITSPERSAMPLE, 1);
            TIFFSetField(*tiffile, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
            break;
        case  RED_MODE:
        case  GREEN_MODE:
        case  BLUE_MODE:
            TIFFSetField(*tiffile, TIFFTAG_PHOTOMETRIC, 1);
            TIFFSetField(*tiffile, TIFFTAG_SAMPLESPERPIXEL, 1);
            TIFFSetField(*tiffile, TIFFTAG_BITSPERSAMPLE, 8);
            TIFFSetField(*tiffile, TIFFTAG_COMPRESSION, COMPRESSION_NONE);
            break;
    }
    
}

/*-------------------------------------------------------------*/
/* 
 * closes a tiff file 
 */

void tiff_close(TIFF *tiffile)
{
    TIFFClose(tiffile);   
    DBG(DBG_SCAN, "tiff_close: file saved\n");
}


/*-------------------------------------------------------------*/
void init_image(scan_image **image) 
{
    int i;

    *image = malloc (sizeof(scan_image));

    (*image)->width = 100;
    (*image)->height = 100;
    (*image)->resolution = 100;
    (*image)->from_top = 0;
    (*image)->from_left = 0;
    (*image)->num_colors = 3;
    (*image)->color_mode = RGB_MODE ;

    for (i = 0; i < 3; i++) {
	((*image)->color)[i].gamma = 2.0;
	((*image)->color)[i].contrast = 0;
	((*image)->color)[i].brightness = 0;
	(*image)->color_table[i] = malloc(color_table_size);
    }

    (*image)->skip_green = 0;
    (*image)->skip_blue = 0;

    (*image)->manual_mode = 0;

    (*image)->filename = malloc(100);
    (*image)->filename = "test.tiff";

}

///////////////////////////////////////////////////////////
//
// Initialize some registers in the scanner
//
void init_scanner()
{
	reg5.w = 0x82;
	epp_write_one( reg5.w , 0x05 );
	reg9.w=0x0a;
	epp_write_one( reg9.w , 0x09 );
	reg31.w=0;
	epp_write_one( reg31.w , 0x1f );		// Buttons + ??
}	

//////////////////////////////////////////////////////////
//
// Status polling routine. Always starts with waiting!
// 
//
void pollc(unsigned char val, unsigned char mask)
{
	do
	{
		usleep(5000);
		epp_read_one(&reg12.w,0xc);
	} while((reg12.w & mask)!=val);
}

///////////////////////////////////////////////////////////
//
// set up the number of motor steps to move next time
// one step = 1/600" !!
void set_scan_length(int slength)
{
	reg7.w = slength/256;
	reg8.w = (unsigned char)slength;

	epp_write_one(reg7.w,0x07);		
	epp_write_one(reg8.w,0x08);
}

///////////////////////////////////////////////////////////
//
// 
void set_line_width(int lwidth)
{
	reg22.w = lwidth/256;
	reg23.w = (unsigned char)lwidth;

	epp_write_one(reg22.w,0x16);		
	epp_write_one(reg23.w,0x17);
}

///////////////////////////////////////////////////////////
//
// 
void set_line_start(int lstart)
{
	reg20.w = lstart/256;
	reg21.w = (unsigned char)lstart;

	epp_write_one(reg20.w,0x14);		
	epp_write_one(reg21.w,0x15);
}

////////////////////////////////////////////////////////////
//
//
void set_memory_addr(int addr)
{
	reg14.w = (addr>>16) & 0x1f;
	reg15.w = addr>>8;
	reg16.w = addr;
	epp_write_one(reg14.w , 0x0e);
	epp_write_one(reg15.w , 0x0f);
	epp_write_one(reg16.w , 0x10);
}
	
///////////////////////////////////////////////////////////
//
// motor trigger period in units of 0.5 usec
//
void set_step_time(int units)
{
	reg10.w = units/256;
	reg11.w = (unsigned char)units;

	epp_write_one(reg10.w,0x0a);		
	epp_write_one(reg11.w,0x0b);
}

///////////////////////////////////////////////////////////
//
// get FIFO counter
//
int get_FIFO_count(void)
{
	epp_read_one(&reg18.w , 0x12);
	epp_read_one(&reg19.w , 0x13);
	return reg18.w * 256 + reg19.w;
} 

///////////////////////////////////////////////////////////
//
// set resolution and reset FIFO
//
void set_resolution(int res)
{
unsigned int pa,pb,pc;

// first pick a prescaler value pc such that pc=300/ydpi and pc<=4

	pc=300/res;
	if(pc>4) pc=4;
	switch ( pc )
	{
		case 4: pa=0xed; break;				// *4
		case 3: pa=0xad; break;				// *3
		case 2: pa=0x6d; break;				// *2
		default:  	pa=0x2d;			// *1
	}

// put high nibbles of x and y resolutions in reg 0x1b

	reg29.w = pc = res * pc;

	pb=( res/256 )<<4;					// x in high nibble
	reg27.w = pb|( pc>>8 );					// merge y

// send resolution to the scanner and do a FIFO reset

	epp_write_one(pa , 0x03);				//
	reg3.w=pa|2;						// turn off FIFO reset
	epp_write_one( reg3.w , 0x03);				//
		
	epp_write_one(reg27.w , 0x1b);				//
	epp_write_one(res , 0x1c);				// dots per inch in x direction ( lsb )
	epp_write_one(reg29.w , 0x1d);				// dots per inch in y direction ( lsb ) * prescaler

//printf("03,1b,1c,1d= %x %x %x %x\n",reg3.w,reg27.w,res,reg29.w);
	
}

///////////////////////////////////////////////////////////
//
// set up register 26 
// "Buffer Bank R/W Difference Motor Start & Stop - R/W"
//
void set_REG26(unsigned char val )
{
	reg26.w=val;
	epp_write_one( reg26.w , 0x1a );
}

///////////////////////////////////////////////////////////
//
// Return the lamp to its home position
//
void go_home(void)
{
	epp_read_one(&reg12.w,0xc);
	if( reg12.r.HomeSensor ) return;

	set_step_time(0x0680);
	epp_write_one(0x08,0x1a);	// test: set higher speed!
	epp_write_one(0x14,0x19);	// test: set higher speed!
	
	epp_write_one(0x03,0x06);	
	epp_write_one(0x17,0x06);
	pollc(0x04,0x04);
}

/////////////////////////////////////////
//
// send one parameter value to the adc
// only bits 13..0 are used. Reg# in upper byte.
//
// don't care about shadow register reg13 here
//
void adc_write(int value)
{
int temp,bit;

	epp_write_one(0x80,0x0d);			// WmVsmp = 4

	for(bit=13;bit>-1;bit--)
	{
		if(value&(1<<bit)) temp=0x82;
			else temp=0x80;
		epp_write_one(temp,0x0d);
		epp_write_one(temp|0x04,0x0d);		// clk hi
		epp_write_one(temp,0x0d);		// clk lo
	}
	epp_write_one(temp|0x08,0x0d);			// cs hi to update the selected register
	epp_write_one(0x80,0x0d);			// cs lo again

	epp_write_one(0x00,0x0d);			// 0x20 WmVsmp = 1
	
}

///////////////////////////////////////////////////////////
//
// set adc offset
// xoff = -255 .. +255
//
void set_adc_offset(int roff, int goff, int boff)
{
//printf("off= %i %i %i  ",roff,goff,boff);
	if(roff&0x100) adc_write( 0x2401); else adc_write( 0x2400);
	if(goff&0x100) adc_write( 0x2501); else adc_write( 0x2500);
	if(boff&0x100) adc_write( 0x2601); else adc_write( 0x2600);
	if(roff&0x100) roff=-roff;
	if(goff&0x100) goff=-goff;
	if(boff&0x100) boff=-boff;
//printf("  set= %i %i %i\n",roff,goff,boff);

	adc_write( 0x2000+(roff&0xff) );
	adc_write( 0x2100+(goff&0xff) );
	adc_write( 0x2200+(boff&0xff) );
}

///////////////////////////////////////////////////////////
//
// set preamplifier gain
//
// Typical Gain = 0.5 + ( Code * 0.25 )
//
void set_adc_gain(int rgain, int ggain, int bgain)
{
	adc_write( 0x2800+(rgain&0x1f) );
	adc_write( 0x2900+(ggain&0x1f) );
	adc_write( 0x2a00+(bgain&0x1f) );
}

////////////////////////////////////////////////////////////
//
// read adc to get average values
//
void read_avg( int *ravg, int *gavg, int *bavg )
{
int i,j;
unsigned char back;
	*ravg=0;
	*gavg=0;
	*bavg=0;

	pollc(0x10,0x10);					// wait for FIFO status
	for(i=0;i<7;i++) 
	{
		do epp_read_one(&back,0x13); while(back<4);	// wait for a minimum of 4 pixels in FIFO ?
//		epp_set_mode(0x11);
		for(j=0;j<4;j++)
		{
			epp_read_one(&back,0x11); *ravg=*ravg+back;	//pgmin();
			epp_read_one(&back,0x11); *gavg=*gavg+back;	//pgmin();
			epp_read_one(&back,0x11); *bavg=*bavg+back;	//pgmin();

//printf(" rgb= %i %i %i\n", *ravg,*gavg,*bavg);

		}
	}
	*ravg=*ravg/28;
	*gavg=*gavg/28;
	*bavg=*bavg/28;
//printf("read_avg1 rgb= %i %i %i\n", *ravg,*gavg,*bavg);
}

/////////////////////////////////////////////////////////////
//
// get an average color reading
// the values do not depend on exposure??
//
void get_avg_reading( int *red, int *green, int *blue )
{

	reg27.w = 0x11;			// preset to 300 dpi
	reg28.w = 0x2c;
	reg29.w = 0x2c;
	epp_write_one(reg27.w , 0x1b);	
	epp_write_one(reg28.w , 0x1c);
	epp_write_one(reg29.w , 0x1d);

	reg3.w = 0x2f;
	epp_write_one(reg3.w , 0x03);

	set_line_start(0x0006);

	set_line_width(0x0007);		// ???

	set_scan_length(0x001e);

	reg26.w = 8;
	reg4.w  = 0x3e;
	epp_write_one(reg26.w , 0x1a);
	epp_write_one(reg4.w , 0x04);

	reg6.w = 0x62;
	epp_write_one(reg6.w , 0x06);

	set_step_time(0x15ea);

	reg6.r.MotorPower = 1;
	epp_write_one(reg6.w , 0x06);

	reg9.w = 0x0a;
	epp_write_one(reg9.w , 0x09);

	reg6.w = 0x72;
	epp_write_one(reg6.w , 0x06);

	reg31.w = 0;
	reg25.w = 0x10;
	epp_write_one(reg31.w , 0x1f);
	epp_write_one(reg25.w , 0x19);

	read_avg(red, green, blue);
}
	
 

///////////////////////////////////////////
//
// adjust offset to get a reading in the range 1 to 10
//
//
void adj_offset(void)
{
int offsetr=0;
int offsetg=0;
int offsetb=0;
int mask,rtmp,gtmp,btmp;

	set_adc_offset(offsetr, offsetg, offsetb);

	for(mask=0x80;mask!=0x00;mask=mask>>1)
	{
		epp_write_one(0x62,6);	//0x62
		epp_write_one(0x2d,3);	// FIFO reset 300 dpi?
		epp_write_one(0x2f,3);	// 

		rtmp=gtmp=btmp=0;
		get_avg_reading(&rtmp,&gtmp,&btmp);

		if(rtmp<0x01) offsetr=offsetr-mask;
		else if(rtmp>0x0a) offsetr=offsetr+mask;

		if(gtmp<0x01) offsetg=offsetg-mask;
		else if(gtmp>0x0a) offsetg=offsetg+mask;

		if(btmp<0x01) offsetb=offsetb-mask;
		else if(btmp>0x0a) offsetb=offsetb+mask;

		set_adc_offset(offsetr, offsetg, offsetb);

	DBG(DBG_MEDIUM,"Read: %i %i %i Offsetregs set to: %2x %2x %2x\n", rtmp,gtmp,btmp,offsetr,offsetg,offsetb);
	}
}
///////////////////////////////////////////////////////////////

void readscandata(scan_image *image, int mode)
{

    TIFF *tiffile;
//    struct timeval old_time, new_time;
    unsigned char *data;
    int y, i, j, pos; 				// timeout; /* check for delay */
    unsigned char back;
//    long long timer;
//FILE *testfile;    
    int width, height, num_colors, resolution;
    int qwerty,blocks,bufptr; 			//*gk
    int qw;
//unsigned char mydata[1000000];

//printf("reading scan data..\n"); //*gk


    width = image->width;
    height = image->height;
    num_colors = image->num_colors;
    resolution = image->resolution;

    DBG(DBG_SCAN, "read_scan_data: begin: %ix%i\n", width, height); 

    DBG(DBG_SCAN, "expecting %i bytes of data\n",width*height*num_colors);
    data = calloc(width * num_colors * num_buffers, sizeof(unsigned char));   

    tiff_open(&tiffile, image);
    blocks=width*num_colors;
//printf("line= %i\n",blocks);

    pos=0;
    y=0;
    qw=0;
    bufptr=0;

    while(y<height){
	qwerty = get_FIFO_count();
	while (!qwerty) {
		/* check for data */
		qwerty = get_FIFO_count();
        }


//printf("              %i     \r",qwerty);

	/* read data */

// new code here:
	if ((qw == qwerty) || (qwerty >= blocks))
        {
          i = blocks - pos;     //free in buffer
          if (i > qwerty)
            i = qwerty;
//for(j=0;j<i;j++) data[pos+j]=0x80; //
          epp_read (data + pos, i, 0x11);
          pos += i;
        }
	else
          qw = qwerty;

	if(pos==blocks){

// convert to GRAY_MODE
		if(mode==GRAY_MODE)
		{	

			j=0;
			for(i=0;i<blocks;i=i+3)
			{
				data[j] = ( 30 * data[ i  ] +
					    59 * data[ i+1] +
					    11 * data[ i+2] ) / 100  ;
				j++;
			}
		}

    		if ( mode==BW_MODE )
    		{
				
        		for (i = 0; i < (blocks / 8) ; i++)
        		{
            			qwerty= 0;
            			for (j = 0; j < 8 ; j++)
            			{
                			qwerty = qwerty << 1;
                			if ( data[i*8+j]  > 0x7f) qwerty = qwerty | 0x01;
            			}
            			data[i] = qwerty  & 0xff;
			}
        	}


// test 
/*

		for(i=0;i<width*num_colors;i=i+3)
		{
			if(y%3    ) data[i+1] = 0; //data[bufptr + i+3+0];
			if((y+1)%3) data[i+2] = 0; //data[bufptr + i+3+2];
			if((y+2)%3) data[i+3] = 0; //data[bufptr + i+3+2];
		}

*/

// overwrite with testpattern of black vertical lines
/*

		for(i=0;i<width*num_colors;i=i+3)
		{
			data[i  ] = i%9 ? 0xff : 0x00 ;
			data[i+1] = i%9 ? 0xff : 0x00 ;
			data[i+2] = i%9 ? 0xff : 0x00 ;
		}
*/

    		if(y) TIFFWriteScanline(tiffile, data, y-1, 0);		// skip first line, it is often discolored

		y++;
		pos=0;
//printf("saved %i\n",qwerty);
	};

    };
    tiff_close(tiffile);

// read some extra data, just for fun

	blocks=0;
	do {
		qwerty = get_FIFO_count();

		if ( qwerty )
		{
//			epp_set_mode(0x11);
//			();
			epp_read_one(&back,0x11);
			blocks++;
		};
	} while( qwerty );
//printf("Got %i extra bytes!\n",blocks);

}
/*-------------------------------------------------------------*/


/*-------------------------------------------------------------*/
/*
 * ugly, ugly ....
 */
void scan(scan_image *image, int mode)
{
int i;
unsigned char *eppdata, *eppindex;    //*pp

// wait for button if in manual mode
	reg31.w = 0;
	epp_write_one(reg31.w , 0x1f);	// clear buttons
	if(image->manual_mode)  
		do
		{ 
			epp_read_one(&reg31.w , 0x1f);
//			printf("0x1f=%02x\n",back);
		} while( reg31.r.Button1 == 0 );

	go_home();					// just in case we are not in the home position

	set_resolution( 300 );

	epp_write_one(0x3e,0x4);			// 0x3e set scan mode ...
	epp_write_one(0x03,6);				// 0x80

//
// set up the registers in the ADC to 12-bit mode and gain=1 ( among other things )
//
	adc_write(0x0510);				// 0x0510 MODE 12
	reg5.w = 0x82;					// Adc1210=1   Sensor = 2
	epp_write_one(reg5.w , 5);
	epp_write_one(0x14,0x19);
	adc_write(0x0103);				// 0x0103 CDS, ENADC ( = default )
	adc_write(0x0204);				// 0x0204 INVOP
	adc_write(0x0312);				// 0x0322 CHAN=0,0 CDSREF=1,0 RLC=1,0
	adc_write(0x2700);				// 0x2700 DAC SIGNS = 0,0,0
	adc_write(0x2b02);				// 0x2b02 rgain=ggain=bgain= *1

	epp_write_one(0x03,6);				// 0x80 0x03

// fill the x-map in 0x9000 with zeroes

	set_memory_addr(0x09000);

	for(i=0;i<0x1000*3*3;i++) epp_write_one(0x00,0x11);	// slow


// fill the area 0x3000..0x8fff with 0x8000 ??

	set_memory_addr(0x03000);

	for(i=0;i<0x1000*3;i++) { epp_write_one(0x00,0x11); epp_write_one(0x00,0x11); }	// slow


// fill the 12->8 bits maps with a default pattern that gives 1:1 ratio in the low end of the adc range
// and 31:1 ratio in the rest of the adc range

	epp_write_one(0x63,0x06);

	set_memory_addr(0x00000);

	epp_set_mode(0x11);

/*
#ifdef SOFTEPP
	for(i=0;i<0x80;i++) pgmout(i);
	for(i=0x80;i<0x1000;i++) pgmout(i/31+124);

	for(i=0;i<0x80;i++) pgmout(i);
	for(i=0x80;i<0x1000;i++) pgmout(i/31+124);

	for(i=0;i<0x80;i++) pgmout(i);
	for(i=0x80;i<0x1000;i++) pgmout(i/31+124);
#else
	for(i=0;i<0x80;i++) outb( i, port + 4);
	for(i=0x80;i<0x1000;i++) outb(i/31+124, port + 4);

	for(i=0;i<0x80;i++) outb( i, port + 4);
	for(i=0x80;i<0x1000;i++) outb(i/31+124, port + 4);

	for(i=0;i<0x80;i++) outb( i, port + 4);
	for(i=0x80;i<0x1000;i++) outb(i/31+124, port + 4);
#endif
*/
	eppdata=malloc( 4096 );
	eppindex=eppdata;

	for(i=0;i<0x80;i++) *eppindex++ = i;
	for(i=0x80;i<0x1000;i++) *eppindex++ = i/31+123;	// 124

	epp_write( eppdata , 4096 , 0x11 );
	epp_write( eppdata , 4096 , 0x11 );
	epp_write( eppdata , 4096 , 0x11 );

	free( eppdata );

//
// adjust offsets
//
	DBG(DBG_LOW,"presetting gain to unity\n");
	set_adc_gain(2,2,2);		// gain set to unity! ( was also done above )

	DBG(DBG_LOW,"preset done\nadjusting offsets\n");
	adj_offset();

	DBG(DBG_LOW,"offsets done\n");

// do an extra read to see the final result:
{
int red,green,blue;
	red=0;
	green=0;
	blue=0;

	epp_write_one(0x62,6);	
	epp_write_one(0x2d,3);		// FIFO reset	
	epp_write_one(0x2f,3);	
//printf("get avg readings\n");
	get_avg_reading(&red, &green, &blue);
//printf("rgbtot= %i %i %i \n",red,green,blue);
}

	set_resolution(300);

	set_REG26( 0x18 );

	set_line_start(0);			// start at physical line hidden under the cover
	set_line_width(	0x0320 );
	set_scan_length(0x00a0 );


	epp_write_one(0x7e,4);			// monochrome mode
	
	set_step_time(0x15ea);
	
	epp_write_one(0x63,6);	
	epp_write_one(0x73,6);			// creg=0x85 without this
	
	epp_write_one(0x10,0x19);


	pollc(0x10,0x10);			// test on FIFO empty bit
//
//
// here we are to find out the location of the calibration area and hence the window
//
// 

{
int i,j,k,w,ystart;
unsigned char back;

	ystart=0;

	for(j=0;j<50;j++)				// should be 200.
	{
		do
		{
		} while( get_FIFO_count() < 205 );

//		epp_set_mode(0x11);
		for(k=0;k<200;k++)
		{
			epp_read_one(&back,0x11);	// pgmin();		// skip over the x-calibration strip		
		}
		w=0;
		for(i=0;i<3;i++)
		{
			do
			{
			} while( get_FIFO_count() < 205 );

//			epp_set_mode(0x11);
			for(k=0;k<200;k++)
			{
				epp_read_one(&back,0x11); w=w+back;	//pgmin();		
			}
		}

		if(ystart==0){ if(w>0x78*600) ystart=j; };
//printf("w= %x\n",w);	

	}
//printf("ystart= %x\n",ystart);

}

	epp_write_one(0x62,6);	
	epp_write_one(0x14,0x19);	
	epp_write_one(0x62,6);	
	epp_write_one(0x81,0x0c);		// lamp on + motor move
	epp_write_one(0x14,0x19);

	set_step_time(0x07d0);

	epp_write_one(0xef,3);			// ScanSpeed = dpi prescaler

	epp_write_one(0x63,6);			//
//	epp_write_one(0x77,6);			// go home to the optoswitch
	epp_write_one(0x95,6);			// go home to the optoswitch


// poll c5 missing here ???? ----------------------
//	pollc(0xc5,0xff);


	epp_write_one(0x63,6);
	epp_write_one(0x85,0x0c);		// lamp on

	set_step_time(0x07d0);

	epp_write_one(0x4f,0x03);
	epp_write_one(0x63,6);

	set_scan_length(0x006e);

	epp_write_one(0x63,6);
	epp_write_one(0x6b,6);			// move to the white area


//	pollc(0x89,0x8f);			// no polling here ???

	set_resolution(300);

	set_REG26( 0x38 );			// "Buffer Bank R/W Difference Motor Start & Stop"

	set_line_start(0x0084);
	set_line_width(0x09f6);			// 0x09f6 = 2550 = 8.5" @ 300dpi	
	set_scan_length(0x0028);

	epp_write_one(0x3e,4);	
	
	set_step_time(0x15ea);
	
	epp_write_one(0x63,6);	
	epp_write_one(0x73,6);			// creg=0x85 without this	

	epp_write_one(0x10,0x19);

	pollc(0x10,0x10);			// wait for reg c=0xd1

{
int j,q,redacc,greenacc,blueacc;
unsigned char back1,back,jox;
double rgain,ggain,bgain;

	while( get_FIFO_count() < 256);

//
// read up to 306000 bytes here to find max reading for each color
// THIS POLLING DOES NOT WORK

	redacc=greenacc=blueacc=0;
	q=0;
	do
	{
		back=get_FIFO_count();

		epp_set_mode(0x11);

		if( back )
		{
			if(back > 255) j=256; else j=back;
			for(i=0;i<j;i++)
			{
				epp_read_one(&jox,0x11); if(jox>redacc) redacc=jox; 	//jox=pgmin();
				epp_read_one(&jox,0x11); if(jox>greenacc) greenacc=jox; 	//jox=pgmin();
				epp_read_one(&jox,0x11); if(jox>blueacc) blueacc=jox; 	//jox=pgmin();
				q++;
			};
		};
	} while ( back );

//printf(" no of pixels read = %i, rgbmax= %i %i %i\n",q,redacc,greenacc,blueacc);

//
// from the read above we can set max gain for each color
// TAKE CARE OF OFFSETS HERE !
//

	redmax=rgain=31*(redacc-123);	// 124
	rgain=4000/rgain;
	redacc=(int) ( 4*(rgain-0.5) );
	redmax=(redmax*(redacc+2))/4;

	greenmax=ggain=31*(greenacc-123); // 124
	ggain=4000/ggain;
	greenacc=(int) ( 4*(ggain-0.5) );
	greenmax=(greenmax*(greenacc+2))/4;

	bluemax=bgain=31*(blueacc-123);		// 124
	bgain=4000/bgain;
	blueacc=(int) ( 4*(bgain-0.5) );
	bluemax=(bluemax*(blueacc+2))/4;

	redgain=redacc;
	greengain=greenacc;
	bluegain=blueacc;

	set_adc_gain( redgain, greengain, bluegain );

//printf(" rgb gaincodes = %i %i %i \n",redacc,greenacc,blueacc);

//
// try out a new set of offsets for the current gain settings
//
	adj_offset();

// do an extra read to see the final result:
{
int red,green,blue;
	red=0;
	green=0;
	blue=0;

	epp_write_one(0x62,6);	
	epp_write_one(0x2d,3);			// reset FIFO
	epp_write_one(0x2f,3);	

	get_avg_reading(&red, &green, &blue);
//printf("final offsets = %i %i %i \n",red,green,blue);
	redmin=red;
	greenmin=green;
	bluemin=blue;

}

	epp_write_one(0x91,0x0c);

	set_step_time(0x07d0);			// set fast

	epp_write_one(0x4f,0x03);

	set_scan_length(0x0028);

	epp_write_one(0x6f,0x06);		// reverse move

	pollc(0x99,0xff);			// wait for reg c=0x99


	set_resolution(300);

	set_REG26( 0x38 );			// "Buffer Bank R/W Difference Motor Start & Stop"

	set_line_start(0);			// start scan from hidden physical linestart
	set_line_width(2750);			//  2550 = 8.5" @ 300 dpi ( scanwidth ) + 200 for hidden margin	
	set_scan_length(0x00a0);		// 0x50 = 80 lines ( scanlength )

	epp_write_one(0x3e,4);	
	
	set_step_time(0x15ea);
		
	epp_write_one(0x63,6);	
	epp_write_one(0x73,6);			// creg=0x85 without this	

	epp_write_one(0x10,0x19);


// temporary code for test of calibration
//
{
int line;
	whiteline = calloc(2750*3, sizeof(int));   

	pollc(0x10,0x10);			// wait for reg c=0xd1

	do {epp_read_one(&back1,0x13);} while(back1==0);

	for(line=0;line<10;line++)
	{
		while( get_FIFO_count() < 2750 );		// wait for 2750 pixels
		
		for(i=0;i<2750*3;i=i+3)

		{
			epp_read_one(&jox,0x11);
			whiteline[i]=whiteline[i]+jox;
			epp_read_one(&jox,0x11);
			whiteline[i+1]=whiteline[i+1]+jox;
			epp_read_one(&jox,0x11);
			whiteline[i+2]=whiteline[i+2]+jox;
		};

	};

	for(i=0;i<2750*3;i++) whiteline[i]=whiteline[i]/10;

// find the highest reading for each color

	redlow=greenlow=bluelow=0;
	for(i=0;i<2750*3;i=i+3)
	{
		if(whiteline[i]>redlow) redlow=whiteline[i];
		if(whiteline[i+1]>greenlow) greenlow=whiteline[i+1];
		if(whiteline[i+2]>bluelow) bluelow=whiteline[i+2];
	}
//printf("rgb white low= %2x %2x %2x\n",redlow,greenlow,bluelow);
}


////////////////////////////////////

//printf("sending x-map to 0x9000\n");


// load the x-direction map
// always using 300dpi

	epp_write_one(0x63,0x06);

	set_memory_addr(0x09000);

//	epp_set_mode(0x11);

	redmax=32*(redlow-124);		// 31 ??
	greenmax=32*(greenlow-124);
	bluemax=32*(bluelow-124);

back=0;	// value depends on scanner model ? =118 for 600p?
    eppdata = calloc (3 * 3 * 4096, sizeof (unsigned char));
    eppindex = eppdata + back;
    for (i = 0; i < 2750; i++)
      {
        *eppindex = ((redlow - whiteline[i * 3]) * 3);
        eppindex++;
      };

    eppindex = eppdata + 3 * 4096 + back;
    for (i = 0; i < 2750; i++)
      {
        *eppindex = ((greenlow - whiteline[i * 3 + 1]) * 3);
        eppindex++;
      };

    eppindex = eppdata + 2 * 3 * 4096 +back;
    for (i = 0; i < 2750; i++)
      {
        *eppindex = ((bluelow - whiteline[i * 3 + 2]) * 3);
        eppindex++;
      };

    epp_write (eppdata, 3 * 3 * 4096, 0x11);	////*gk 3*3*4096!
    free (eppdata);

//
// send something to 0x3000..
//

//printf("sending 8000 8000 ...\n");

	set_memory_addr(0x03000);
		
    eppindex = eppdata = calloc (2 * 3 * 4096, sizeof (unsigned char));
    for (i = 0; i < 24576; i += 2)
      {
        *eppindex = 0x80;
        eppindex += 2;
      }
    epp_write (eppdata, 2 * 3 * 4096, 0x11);
    free (eppdata);

	epp_write_one(0x14,0x19);

	set_step_time(0x07d0);

// here we are to add the y-offset found above, use 100 to make sure we don't start outside window

	set_scan_length(100+image->from_top*600);		// 100 steps = 100/600 = 1/6 "

	epp_write_one(0x6b,0x06);				// move to start of scanarea

	calc_color_table(image);				// set up the final color-matrix

//  WIN driver does pollc for 81h here  //pp
	pollc(0x08,0x08);					// wait for reg c=0x89 ??

// set up memory pointer and send color matrix

	set_memory_addr(0x00000);

/* why do it this way ?
    eppindex = eppdata = calloc (3 * 4096, sizeof (unsigned char));
    for (i = 0; i < 3; i++)
      for (j = 0; j < 4096; j++)
        {
          *eppindex = ((char) (image->color_table[i][j]));
          eppindex++;
        }
    epp_write (eppdata, 3 * 4096, 0x11);
    free (eppdata);
*/


//non-replaced code here
	epp_set_mode(0x11);
	for(i=0;i<3;i++) epp_write( image->color_table[i] , 4096 , 0x11 ); 
// end of non-replaced code

/*	
	for(i=0;i<3;i++)
	{
		for(j=0;j<4096;j++)
			printf(" %3i",(image->color_table[i][j]));
		printf("\n");
	}
*/
/*
	for(i=0;i<4096-8;i=i+8)
	{
		printf("\n%3X",i);
		for(j=0;j<8;j++)
			printf(" %2X%2x%2X",image->color_table[0][i+j],image->color_table[1][i+j],image->color_table[2][i+j]);
	}
*/

// do the real scan here
// 
{ 
int height,width,fromleft;


	fromleft=(image->from_left)*300+132;		// add some value from above to skip over hidden margin TO BE FIXED
	width=image->width;				// get width in pixels

	height=(image->height)*600/(image->resolution);	// calculate number of steps at 600 dpi for this distance
//	height=height+96-height%96;			// round up to even multiple of 96
	height=(height/96+1)*96+100; 			// and add some extra!

	DBG(DBG_SCAN," scanning wxh = %i x %i\n",width,height);
	DBG(DBG_SCAN," height hi lo = %x  %x (hex)\n", height/256,(unsigned char)(height&0xff) );
//printf(" width  hi lo = %x  %x (hex)\n",width/256,(unsigned char)width);



	set_resolution(image->resolution);

	set_line_start(fromleft);

	set_line_width(width);

	set_scan_length(height);

	if(image->color_mode==BW_MODE) reg4.w = 0x7e;		// set monochrome 
		else reg4.w = 0x3e;				// rgb mode

	epp_write_one(reg4.w , 0x04);				// scan mode ..

//*pp primax 600p software at this point read one byt from 0dh (in most casses 0x31h
//    is return value, and next epp_write to 0dh is performad with value 0x35h 
//*gk this sets the shift clock for the adc serial line

	if(image->resolution>150) j=2;
		else
		if(image->resolution<101) j=0; else j=1;
	
	switch(j)
	{
		case 2:		// >150 dpi
		{
			epp_write_one( 0x18 , 0x1a );
			set_step_time( 0x15ea );
			epp_write_one( 0x63 , 6 );
			epp_write_one( 0x73 , 6 );
			epp_write_one( 0x10 , 0x19 );
		} break;
	
		case 1:		// 101 .. 150 dpi dpi
		{
			epp_write_one( 0x8 , 0x1a );
			epp_write_one( 0x23 , 6 );
			set_step_time( 0x0af5 );
			epp_write_one( 0x23 , 6 );
			epp_write_one( 0x33 , 6 );
			epp_write_one( 0x14 , 0x19 );
		} break;
	
		default:	// <=100 dpi
		{
			epp_write_one( 0x8 , 0x1a );
			set_step_time( 0x07d0 );
			epp_write_one( 0x03 , 6 );
			epp_write_one( 0x13 , 6 );
			epp_write_one( 0x14 , 0x19 );
		}	
	}
}

}

	readscandata(image, image->color_mode);

//
// return to start position
//

//*  usleep (100000);              //*pp wait before reverse move 
	set_step_time(0x0680);
	epp_write_one(0x08,0x1a);	// test: set higher speed!
	epp_write_one(0x14,0x19);	// test: set higher speed!
	
	epp_write_one(0x03,0x06);	// don't use go_home() here	
	epp_write_one(0x17,0x06);	// it will hang until the lamp is back home

}



/*-------------------------------------------------------------*/
/* 
 * parse command line input
 */
void parse_options(scan_image *image, int argc, char **argv)
{
    int index = 0;
    int c, switch_lamp = 0xFF;
    char *tmp;
    unsigned char back, status;
    double width = 0, height = 0;

    static struct option options[] = {
	{"help",   0, 0, 'h'},
	{"verbose",2, 0, 'v'},
	{"port",   1, 0, 'P'},
	{"dim",    1, 0, 'd'},
	{"res",    1, 0, 'r'},
	{"pos",    1, 0, 'p'},
	{"version",0, 0, 'V'},
	{"color",  1, 0, 'c'},
	{"name",   1, 0, 'n'},
	{"lamp",   1, 0, 'l'},
        {"GRAY",   0, 0, 'G'},
        {"GREY",   0, 0, 'G'},
        {"BW",     0, 0, 'B'},
        {"manual", 0, 0, 'm'},
        {"quiet",  0, 0, 'q'},
	{0,        0, 0,   0}
    };
    
    dbg_level = 0; // DBG_HELPERS;
    DBG(DBG_HELPERS,"parse_options:\n");
    while (1) {
	 c = getopt_long(argc, argv, "hv:P:d:Vp:d:r:c:n:l:GBmq",
			 options, &index);
	 if (c == -1)
	     break;

	 switch (c){
	 case 'P':
	     set_port( strtol (optarg, (char **)NULL, 0x10));
//	     DBG(DBG_HELPERS, "  port:               0x%03x\n", port);
	     break;

	 case 'v':
	     if (!optarg)
		 dbg_level = DBG_HIGH;
	     else
		 dbg_level = strtol (optarg, (char **)NULL, 0x10);
	     DBG(DBG_HELPERS, "  debug level:        0x%02x\n", (unsigned char)dbg_level);
	     break;

	 case 'V':
	     printf("pxscan version ");
	     printf(VERSION);
	     printf("\n");
	     exit(0);

	 case 'p':
	     image->from_top = strtod(optarg, &tmp);
	     image->from_left =strtod((char *)tmp + 1, (char **)NULL);
	     DBG(DBG_HELPERS, "  position:         %1.1fx%1.1f inch\n", 
		 image->from_top, image->from_left);
	     break;

	 case 'd':
	     width = strtod(optarg, &tmp);
	     height = strtod((char *)tmp + 1, (char **)NULL);
	     DBG(DBG_HELPERS, "  dimension:        %1.1fx%1.1f inch\n", 
		 width, height);
	     break;

	 case 'r':
	     image->resolution = strtol(optarg, (char **) NULL, 10);
	     DBG(DBG_HELPERS, "  resolution:           %i dpi\n", 
		 image->resolution);
	     break;

	 case 'c':
	     image->color[0].contrast = strtod(optarg, &tmp);
	     image->color[0].brightness = strtod((char *)tmp + 1, &tmp);
	     image->color[0].gamma = strtod((char *)tmp + 1, (char **)NULL);

	     image->color[1].contrast =  image->color[0].contrast;
	     image->color[1].brightness = image->color[0].brightness;
	     image->color[1].gamma = image->color[0].gamma;

	     image->color[2].contrast =  image->color[0].contrast;
	     image->color[2].brightness = image->color[0].brightness;
	     image->color[2].gamma = image->color[0].gamma;

	  if(!quietflag) printf( "  color:       %1.1f/%1.1f/%1.1f \n", 
		 image->color[0].contrast, 
		 image->color[0].brightness, 
		 image->color[0].gamma);
             break;

	 case 'n':
	     image->filename = optarg;
	     DBG(DBG_HELPERS, "  filename:     %s\n",image->filename); 
//printf("  filename:     %s\n",image->filename);
	     break;

         case 'G':
             image->color_mode = GRAY_MODE;
//printf("Setting GRAY mode\n");
             break;

         case 'B':
             image->color_mode = BW_MODE;
	     image->num_colors = 1;
//printf("Setting BW mode\n");
             break;

          case 'm':
             image->manual_mode = 1;
//printf("Setting manual mode\n");
             break;

	 case 'l':
	     if (strcmp((const char *)optarg, "off"))
		 switch_lamp = 0x20;
	     else
		 switch_lamp = 0x00;
	     DBG(DBG_HELPERS, "  lamp:                 %s\n",optarg); 
	     break;

         case 'q':
	     quietflag=1;
	     break;


	 case 'h':
	     printf("Usage:   primax_scane [OPTIONS]\n");
	     printf("  -h,   --help             shows this message\n");
	     printf("  -v,   --verbose=LEVEL    increase debugging output\n");
	     printf("  -P,   --port=PORT        portnumber of parallel port\n");
	     printf("  -V,   --version          display version\n");
	     printf("  -p    --pos=BXL          position from bottomxleft in inch\n");
	     printf("  -d    --dim=WxH          widthxheigth of scan area in inch\n");
	     printf("  -c    --color=C/B/G      contrast/brightness/gamma\n");
	     printf("  -r    --res=RES          resolution\n");
	     printf("  -n    --filename=NAME    name for tiff file\n");
	     printf("  -l    --lamp=on/off      turn lamp on/off\n");
             printf("  -m    --manual           press button to start scanning\n");
             printf("  -B    --BW               Black/White mode\n");
             printf("  -G    --GREY             Greyscale mode\n");
             printf("  -q    --quiet            quiet mode\n");
	     exit(0);
	 }
    }

	if(image->resolution>300)  { 	if(!quietflag) printf(" Sorry, Resolutions above 300 dpi are not supported \n"); exit(1);}
		else if(image->resolution<2) {	if(!quietflag) printf(" Sorry, Resolutions below 2 dpi are not supported \n"); exit(1);};

    image->width = (int)(image->resolution * width);
    image->height = (int)(image->resolution * height);
     
    DBG(DBG_HELPERS, "  dimension:        %3ix%3i pixel\n", 
	image->width, image->height);

    if (switch_lamp != 0xFF) {
	DBG(DBG_HELPERS, "parse_options:       switching lamp\n");

//printf("switch_lamp=%d\n",switch_lamp);

	epp_on();						// get the printer port
	switcher_online();

// try some warm up stuff here

	if(switch_lamp==0x20)	epp_write_one(0x80,0x0c);	// turn the lamp on
	else	epp_write_one(0x00,0x0c);			// turn the lamp off

	status = epp_read_one(&back, 0x00);			// ASIC-Type
       
       	if(!quietflag) printf("\nASIC: %02X\n",back);

       	status = epp_read_one(&reg24.w , 0x18); 		// Scanner model
	if(!quietflag) printf("scanner model reg= %0x\n",reg24.r.ProductID);

	go_home();						// return the lamp to its home position if nedeed

	switcher_offline();					// deselect the scanner
	epp_off();						// restore the printer port

	if (!status) {
	    switch ( reg24.r.ProductID) {
	    case 0x01:
		if(!quietflag) printf("\nLooks like a Primax Colorado Direct 9600\n");
		break;
            case 0x02:
		if(!quietflag) printf("\nModel: Primax Colorado Direct\n");
                break;
	    case 0x05:
              	if(!quietflag) printf("\nModel: Primax Colorado Direct 600 or 1200p\n");
		break;
	    default:
		if(!quietflag) printf("\nCan not identify this scanner. code: %02x\n", back&0x1f);
	    }

	    exit(0);

	}
	else {
	    if(!quietflag) printf("\n:-(\nScanner did not switch to EPP mode");
	    if(!quietflag) printf(".....not compatible?\n");
	    exit(1);

	}

	switcher_offline();					// deselect the scanner
	epp_off();						// restore the printer port

	exit(0);
    }
    if(!quietflag) printf("pxscan version ");
    if(!quietflag) printf( VERSION );
    if(!quietflag) printf("\n");



}


/*-------------------------------------------------------------*/

int main(int argc, char **argv)
{	
scan_image *image;


	init_image( &image);

    	parse_options(image, argc, argv);

    	epp_on();

	switcher_online();

    	init_scanner();
 
    	scan(image, SCAN);

	switcher_offline();
    	epp_off();

    	free(image);
    	exit(0);

}

/* This stuff is for emacs
 * Local variables:
 * c-basic-offset: 4
 * End:       
 */
